/*
 * attackwatch.c
 *
 * (c) 2001 by Andreas Beck <becka@bedatec.de>
 *
 * preview version - do not distribute
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "config.h"

struct parms {
	int maxprobes;
	char packetlog[1024];
	char mylog[1024];
	char handlerdir[1024];
} parms= {
	MAX_PROBES,
	PACKETLOG,
	MYLOG,
	HANDLERDIR
};

struct mydata {
	FILE *mylog;
	const char *prgnam;
	char hostname[256];
	int	mypid;
} mydata = {
	NULL,
	"",
	"",
	0
};


static int init_mydata(char *prgnam) {

	int logfd;
	
	mydata.prgnam=prgnam;
	if (strrchr(mydata.prgnam,'/')) mydata.prgnam=strrchr(mydata.prgnam,'/')+1;

	if (-1==(logfd=open(parms.mylog,O_CREAT|O_APPEND|O_WRONLY,S_IRUSR|S_IWUSR))) {
		fprintf(stderr,"Cannot open logfile.Aborting.\n");
		return 1;
	}
	if (NULL==(mydata.mylog=fdopen(logfd,"a"))) {
		close(logfd);
		fprintf(stderr,"Cannot fdopen logfile.Aborting.\n");
		return 1;
	}
	dup2(logfd,1);
	dup2(logfd,2);
	close(0);
	gethostname(mydata.hostname,sizeof(mydata.hostname));
	mydata.mypid=(int)getpid();
	return 0;
}

static int exit_mydata(void) {

	fclose(mydata.mylog);
	return 0;
}

static char *logdate(void) {
	static char timebuf[256];
	struct tm *tm;
	time_t tim;
	time(&tim);
	tm=localtime(&tim);
	strcpy(timebuf,asctime(tm)+4);
	timebuf[15]='\0';
	return timebuf;
}

int dolog(char *format,...) {
	va_list ap;
	int rc;

	va_start(ap,format);

	fprintf(mydata.mylog,"%s %s %s[%d]: ",
		logdate(),mydata.hostname,mydata.prgnam,(int)mydata.mypid);

	rc=vfprintf(mydata.mylog,format,ap);
	fprintf(mydata.mylog,"\n");
	fflush(mydata.mylog);

	va_end(ap);
	return rc;
}

int call_handler(char *name,char *offending_ip) {

	char buffer[1024];
	
	strcpy(buffer,parms.handlerdir);
	strcat(buffer,"/");
	strcat(buffer,name);
	if (access(buffer,X_OK)) return 1;
	dolog("%s %s.",name,offending_ip);
	switch(fork()) {
		case -1:dolog("Could not fork().");break;
		case 0:	/* child */
			execl(buffer,name,offending_ip,NULL);
			exit(1);
			break;
		default:	/* parent */
			wait(NULL);
			break;
	}
	
	return 0;
}

struct logstat {
	off_t	lastsize;
} logstat= {
	0
};

int dowatch(void) {

	FILE *plog;
	struct stat sbuf;
	char inbuffer[1024];
	char lastsrc[32]="";
	int lastcnt=0;
	int len;
	char *hlp,*chain,*jump,*device,*srcip,*dstip;
	int proto,srcport,dstport;
	
	do {
		/* Wait until file has grown */
		do {
			sleep(1);

			if (stat(parms.packetlog,&sbuf)) {
				/* The file is not there - we just assume has been
				   moved temporarily. */
				continue;
			}

			if (sbuf.st_size<logstat.lastsize) {
				/* Oops - the file shrunk ? We assume it was truncated.
				 */
				logstat.lastsize=0;
				dolog("logfile was truncated");
			}

		} while (sbuf.st_size==logstat.lastsize);

		if (NULL==(plog=fopen(parms.packetlog,"r"))) {
			/* Hmm - the file should be there - oh well bad timing
			 * we just retry.
			 */
			continue;
		}

		if (fseek(plog,logstat.lastsize,SEEK_SET)) {
			/* We should be able to seek there ... strange.
			 */
			fclose(plog);
			continue;
		}
		while(NULL!=fgets(inbuffer,sizeof(inbuffer),plog)) {
			len=strlen(inbuffer);
			if (len>0&&inbuffer[len-1]!='\n') {
				/* we read an incomplete line wait for it to finish.
				 */
				break;
			} else inbuffer[len]='\0';
			logstat.lastsize=ftell(plog);
			if (NULL==(hlp=strstr(inbuffer," kernel: Packet log: "))) {
				/* This is not a Packet log line.
				 */
				continue;
			}
			hlp+=21;/* strlen(" kernel: Packet log: ") */
		/* Now break down the log line. */
		
			chain=hlp;
			if (NULL==(hlp=strchr(hlp,' '))) goto illegal_line;
			*hlp='\0';jump=hlp+1;
			if (NULL==(hlp=strchr(hlp+1,' '))) goto illegal_line;
			*hlp='\0';device=hlp+1;
			if (NULL==(hlp=strchr(hlp+1,' '))) goto illegal_line;
			*hlp='\0';
			if (NULL==(hlp=strchr(hlp+1,'='))) goto illegal_line;
			proto=atoi(hlp+1);
			if (NULL==(hlp=strchr(hlp+1,' '))) goto illegal_line;
			srcip=hlp+1;
			if (NULL==(hlp=strchr(hlp+1,':'))) goto illegal_line;
			*hlp='\0';srcport=atoi(hlp+1);
			if (NULL==(hlp=strchr(hlp+1,' '))) goto illegal_line;
			dstip=hlp+1;
			if (NULL==(hlp=strchr(hlp+1,':'))) goto illegal_line;
			*hlp='\0';dstport=atoi(hlp+1);
			if (NULL==(hlp=strchr(hlp+1,' '))) goto illegal_line;
			
			/* O.K. now for the rules.
			 */
			
			/* The blocking chain will never emit logs unless the
			 * sender hit the unblocking rule that should be on
			 * top of this chain. If we get anything from there,
			 * we request to unblock the IP.
			 */
			if (0==strcmp(chain,"block")) {
				call_handler("unblock",srcip);
			} else {
				if (0==strcmp(lastsrc,srcip))	{
					if (++lastcnt>=parms.maxprobes) {
						call_handler("block",srcip);
						/* we blocked it. Ignore the next 10 entries for that host
						 * They will be reported as continuing attacks.
						 */
						lastcnt=-10;
					} else {
						call_handler("attack_cont",srcip);
					}
				} else {
					strcpy(lastsrc,srcip);
					lastcnt=1;
					call_handler("attack_new",srcip);
				}
			}
			continue;
			illegal_line:
			dolog("Got illegal log line:%s",inbuffer);
		}
		fclose(plog);
	} while(1);
	return 0;
}

int daemonize(void) {
	switch(fork()) {
		case -1:
			fprintf(stderr,"Cannot fork to become daemon.");
			return -1;
		case 0: 
			setsid();
			return 0;
		default:
			return 1;
	}
}

int main(int argc, char **argv) {

	if (argc==2&&0==strcmp(argv[1],"-D")) {
		switch(daemonize()) {
			case 0: break;
			case 1: return 0;
			default: return 1;
		}
	}
	if (init_mydata(argv[0])) return 1;
	dolog("Starting service");

	dowatch();

	dolog("Stopping Service");
	exit_mydata();
	return 0;
}
